# Generated by Django 2.2 on 2020-01-26 17:47

from django.db import migrations


def backward_weirdly_named(david, test):
    pass


def forward_weird_name(arg1, arg2):
    pass


class Migration(migrations.Migration):

    dependencies = [("app_data_migrations", "0002_missing_reverse")]

    operations = [migrations.RunPython(forward_weird_name, backward_weirdly_named)]
